// (C) Copyright 2002-2007 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//

//-----------------------------------------------------------------------------
//----- acrxEntryPoint.cpp
//-----------------------------------------------------------------------------
#include "StdAfx.h"
#include "resource.h"

#include "AcPointCloud.h"

//-----------------------------------------------------------------------------
#define szRDS _RXST("")

//-----------------------------------------------------------------------------
//----- ObjectARX EntryPoint
class CAsdkPointCloudApp : public AcRxArxApp {

public:
	CAsdkPointCloudApp () : AcRxArxApp () {}

	virtual AcRx::AppRetCode On_kInitAppMsg (void *pkt) {

		AcRx::AppRetCode retCode =AcRxArxApp::On_kInitAppMsg (pkt) ;
		
		bool ret = acrxLoadModule(L"AcDbPointCloudObj.dbx", false); 

		return (retCode) ;
	}

	virtual AcRx::AppRetCode On_kUnloadAppMsg (void *pkt) {
		AcRx::AppRetCode retCode =AcRxArxApp::On_kUnloadAppMsg (pkt) ;

		return (retCode) ;
	}

	virtual void RegisterServerComponents () {
	}
	static AcDbObjectId selectPointCloud()
	{
		ads_point pt;
		ads_name name;
		if (acedEntSel(L"Select a point cloud", name, pt) != RTNORM)
			return AcDbObjectId::kNull;

		AcDbObjectId id;
		if (acdbGetObjectId(id, name) != Acad::eOk)
			return AcDbObjectId::kNull;

		return id;
	}

	class MyProgress : IAcPcIndexProgress
	{
		/// <summary>
		/// This callback function indicates that the indexing process successes.
		/// </summary>
		virtual void finished(float progress, Acad::ErrorStatus nResult)
		{
		}
	};

	// creates a new point cloud entity based on the las or pcg file that is part of the project folder
	static void AsdkPointCloud_Cloud_Create(void)
	{
		// create a *.pcg file from *.las
		AcString input(L"C:\\Demo Head - Point Cloud.las");
		AcString output(L"C:\\Demo Head - Point Cloud.pcg");
		MyProgress progress;
		// this function can create a pcg file out of a different format, e.g. *.las
		Acad::ErrorStatus err = acpcIndexPointCloud(input, output, (IAcPcIndexProgress*)&progress);
		if (err != Acad::eOk)
			return;

		// use the pcg file to insert a point cloud
		AcDbEntity* pCloud;	
		err = acdbCreatePointCloudEntity(pCloud, output, input, AcGePoint3d(0, 0, 0));
		if (err != Acad::eOk)
			return;

		AcDbDatabase* pDb = acdbHostApplicationServices()->workingDatabase();  
		AcDbBlockTableRecordPointer ptrMS(ACDB_MODEL_SPACE, pDb, AcDb::kForWrite);
		if (ptrMS.openStatus() != Acad::eOk)
			return;

		err = ptrMS->appendAcDbEntity(pCloud);
		pCloud->close(); 
	}

	class MyFilter : IAcPcPointFilter
	{
		virtual void doFilter(const IAcPcDataBuffer& inBuffer, IAcPcDataBuffer& outBuffer)
		{
			bool ret;
			for (DWORD i = 0; i < inBuffer.size(); i++)
			{
				AcPcPointDouble pt;
				ret = inBuffer.doublePoint(i, pt);

				if (pt.m_x > 0)
				{
					pt.m_x += 50;
					pt.m_argb = 0xffff0000; // no effect
					ret = outBuffer.setDoublePoint(i, pt); 
				}
			}
		}
	};

	// adds a filter which is able to change the position and color of any of the points
	// you need to make sure of two things in order to see the color change as well:
	// 1) you need to be in a 3d view (i.e. not 2d wireframe)
	// 2) you need to use hardware acceleration: options > system > performance settings > manual tune > enable hardware acceleration
	static void AsdkPointCloud_Cloud_AddFilter(void)
	{
		AcDbEntityPointer ptrCloud(selectPointCloud(), AcDb::kForWrite); 
		if (ptrCloud.openStatus() != Acad::eOk)
			return;

		static MyFilter filter;
		Acad::ErrorStatus err = acdbModifyPointCloudDataView(ptrCloud, (IAcPcPointFilter*)&filter);
		if (err != Acad::eOk)
			return;
	}

	static void AsdkPointCloud_Cloud_RemoveFilter(void)
	{
		AcDbEntityPointer ptrCloud(selectPointCloud(), AcDb::kForWrite); 
		if (ptrCloud.openStatus() != Acad::eOk)
			return;

		Acad::ErrorStatus err = acdbResetPointCloudDataView(ptrCloud);
		if (err != Acad::eOk)
			return;
	}

	class MyProcessor : public IAcPcPointProcessor 
	{
		/// <summary>Return true to cancel processing.</summary>
		virtual bool cancel()
		{
			return false;
		}

		/// <summary>Called to abort the process</summary>
		virtual void abort()
		{
		}

		/// <summary>Called after the last call to ProcessPoints</summary>
		virtual void finished()
		{
		}

		/// <summary>Process point in buffer. This will be called repeatedly with 
		/// new buffer contents until all points have been processed.</summary>
		virtual bool processPoints()
		{
			AcDbDatabase* pDb = acdbHostApplicationServices()->workingDatabase();  
			AcDbBlockTableRecordPointer ptrMS(ACDB_MODEL_SPACE, pDb, AcDb::kForWrite);
			if (ptrMS.openStatus() != Acad::eOk)
				return false;

      double x, y, z;
      buffer()->offset(x, y, z);
      AcGeVector3d offset(x, y, z);

			for (DWORD i = 1; i < buffer()->size(); i++)
			{
				AcDbObjectPointer<AcDbLine> ptrLine;
				ptrLine.create(); 

				AcPcPointDouble pt;
				buffer()->doublePoint(i - 1, pt);
				ptrLine->setStartPoint(AcGePoint3d(pt.m_x, pt.m_y, pt.m_z) + offset);
				buffer()->doublePoint(i, pt);
				ptrLine->setEndPoint(AcGePoint3d(pt.m_x, pt.m_y, pt.m_z) + offset);
				ptrLine->setColorIndex(2);

				ptrMS->appendAcDbEntity(ptrLine); 
			}

			return true; 
		}

		/// <summary>Filter, if available, to filter points.
		/// Called before processPoints() on the next batch of points.</summary>
		virtual IAcPcPointFilter*  filter()
		{
			return NULL; // or create and use an instance of a filter like MyFilter class
		}
	};

	// this function shows how you can reach the points and do something with them
	// I'll create lines based on the points - this is also a good way to show the cells/regions which the points are organized into
	// To avoid the graphics disappearing every now and then when rotating the view, it's better to disable adaptive degradation:
	// options > system > performance settings > adaptive degradation
	static void AsdkPointCloud_Cloud_ProcessPoints(void)
	{	
		AcDbEntityPointer ptrCloud(selectPointCloud(), AcDb::kForWrite); 
		if (ptrCloud.openStatus() != Acad::eOk)
			return;

    AcDbExtents ext;
    ptrCloud->getGeomExtents(ext);
    ext.set(ext.minPoint(), AcGePoint3d(ext.minPoint().x + ((ext.maxPoint().x - ext.minPoint().x) / 2.0), ext.maxPoint().y, ext.maxPoint().z));

		// you can use either AcViewFrustum or AcDbExtents
		// frustom needs to be defined the following way
		// X = Left/Right
		// Y = Upper/Lower
		// Z = Near/Far
		AcViewFrustum vf;

		vf.setFarLowerLeft (ext.minPoint().x, ext.minPoint().y, ext.maxPoint().z);
		vf.setFarLowerRight(ext.maxPoint().x, ext.minPoint().y, ext.maxPoint().z);
		vf.setFarUpperLeft (ext.minPoint().x, ext.maxPoint().y, ext.maxPoint().z);
		vf.setFarUpperRight(ext.maxPoint().x, ext.maxPoint().y, ext.maxPoint().z);

		vf.setNearLowerLeft (ext.minPoint().x, ext.minPoint().y, ext.minPoint().z);
		vf.setNearLowerRight(ext.maxPoint().x, ext.minPoint().y, ext.minPoint().z);
		vf.setNearUpperLeft (ext.minPoint().x, ext.maxPoint().y, ext.minPoint().z);
		vf.setNearUpperRight(ext.maxPoint().x, ext.maxPoint().y, ext.minPoint().z);

		vf.setNearClip(false);
		vf.setFarClip(false); 

		MyProcessor processor;
		Acad::ErrorStatus err = acdbProcessPointCloudData(ptrCloud, vf, 100, (IAcPcPointProcessor*)&processor); 
		
		// the above code does the same as this
		/*
		MyProcessor processor;
		Acad::ErrorStatus err = acdbProcessPointCloudData(ptrCloud, ext, 100, (IAcPcPointProcessor*)&processor);
		*/
	}
};

//-----------------------------------------------------------------------------
IMPLEMENT_ARX_ENTRYPOINT(CAsdkPointCloudApp)

ACED_ARXCOMMAND_ENTRY_AUTO(CAsdkPointCloudApp, AsdkPointCloud, _Cloud_Create, Cloud_Create, ACRX_CMD_TRANSPARENT, NULL)
ACED_ARXCOMMAND_ENTRY_AUTO(CAsdkPointCloudApp, AsdkPointCloud, _Cloud_AddFilter, Cloud_AddFilter, ACRX_CMD_TRANSPARENT, NULL)
ACED_ARXCOMMAND_ENTRY_AUTO(CAsdkPointCloudApp, AsdkPointCloud, _Cloud_RemoveFilter, Cloud_RemoveFilter, ACRX_CMD_TRANSPARENT, NULL)
ACED_ARXCOMMAND_ENTRY_AUTO(CAsdkPointCloudApp, AsdkPointCloud, _Cloud_ProcessPoints, Cloud_ProcessPoints, ACRX_CMD_TRANSPARENT, NULL)
